from __future__ import division

import random
from warnings import warn

import graph_algorithms as graphalg
import matrix_operations as matop
import numpy as np
from cliquetree import CliqueTree
from graph_algorithms import make_mask_graph
from solvers import (maxdet_complete, grone_order, feasible_interval,
                     integral_function_det, function_det)
from utils import (MAXINT, is_pd, is_symmetric, rearrange, count_nans,
                   get_largest_submatrix, log_subtract, plot_boolean_matrix)
from sklearn.utils import check_random_state


def _oracle(mtx, i, j):
    return mtx[i, j]


def rand_query(partial, mtx, budget, clique_tree=None, random_state=None):
    prng = check_random_state(random_state)
    random.seed(prng.randint(MAXINT))
    if clique_tree is None:
        clqtree = CliqueTree()
        clqtree.from_graph(make_mask_graph(partial))
    else:
        clqtree = clique_tree.copy()
    while budget > 0 and len(clqtree.insertable):
        # e = random.choice(clqtree.insertable, 1)[0]
        pool = [x for x in clqtree.insertable]
        e = pool[prng.choice(len(pool), 1)[0]]
        i, j = e
        clqtree.add_edge(i, j)
        partial[i, j] = partial[j, i] = _oracle(mtx, i, j)
        budget -= 1
    return partial, budget * 2


def greedy_structure(partial, mtx, budget, clique_tree=None):
    if clique_tree is None:
        clqtree = CliqueTree()
        clqtree.from_graph(make_mask_graph(partial))
    else:
        clqtree = clique_tree.copy()
    while budget > 0 and len(clqtree.insertable):
        best = (0, None)
        for e in clqtree.insertable:
            temptree = clqtree.copy()
            temptree.add_edge(*e)
            weight = len(temptree.insertable)
            if weight > best[0]:
                best = (weight, e)
        i, j = best[1]
        clqtree.add_edge(i, j)
        partial[i, j] = partial[j, i] = _oracle(mtx, i, j)
        budget -= 1
    return partial, budget * 2


def greedy_struct_query(partial, mtx, budget, clique_tree=None):
    if clique_tree is None:
        clqtree = CliqueTree()
        clqtree.from_grph(make_mask_graph(partial))
    else:
        clqtree = clique_tree.copy()

    while budget > 0 and len(clqtree.insertable):
        best_option = (0, None)
        for e in clqtree.insertable:
            i, j = e
            temp_clqtree = clqtree.copy()
            temp_clqtree.add_edge(*e)
            candidates = len(temp_clqtree.insertable)
            if candidates >= best_option[0]:
                best_option = (candidates, e)

        i, j = best_option[1]
        clqtree.add_edge(i, j)
        partial[i, j] = partial[j, i] = _oracle(mtx, i, j)
        budget -= 1
    return partial, budget * 2


def grone_query(partial, mtx, budget, sample=False):
    edges = grone_order(partial, sample=sample)[::-1]  # reverse the list and pop() elements
    while budget > 0 and edges:
        e = edges.pop()
        i, j = e
        partial[i, j] = partial[j, i] = _oracle(mtx, i, j)
        budget -= 1
    return partial, budget * 2


def global_impact(pos, partial, retries=10, max_attempts=5, random_state=None):
    i, j = pos
    prng = check_random_state(random_state)
    local_partial = get_largest_submatrix(partial, i, j)
    center, interval = feasible_interval(local_partial)
    partial_cpy = partial.copy()

    partial_cpy[i, j] = partial_cpy[j, i] = center
    maxdet_complete_glb = maxdet_complete(partial_cpy)
    sgn, log_maxdet_global = np.linalg.slogdet(maxdet_complete_glb)

    incr = (interval[1] - interval[0]) * 0.25
    incr = 0

    total_impact = 0
    for retry in range(retries):
        attempts = 1
        partial_cpy = partial.copy()
        rand_fill = prng.uniform(interval[0] + incr, interval[1] - incr)
        partial_cpy[i, j] = partial_cpy[j, i] = rand_fill
        randdet_complete_glb = maxdet_complete(partial_cpy)
        sgn, log_randdet_global = np.linalg.slogdet(randdet_complete_glb)
        while sgn < 0 and attempts < max_attempts:
            partial_cpy = partial.copy()
            rand_fill = prng.uniform(interval[0] + incr, interval[1] - incr)
            partial_cpy[i, j] = partial_cpy[j, i] = rand_fill
            randdet_complete_glb = maxdet_complete(partial_cpy)
            sgn, log_randdet_global = np.linalg.slogdet(randdet_complete_glb)
            max_attempts += 1
        if sgn < 0:
            raise ValueError('Choice for (%d, %d) results in negative determinant' % (i, j))

        impact = log_subtract(log_maxdet_global, log_randdet_global) -\
            log_maxdet_global
        total_impact += impact

    return np.exp(total_impact / retries)


def pre_global_impact(partial, retries=10, max_attempts=5, random_state=None):
    """For each missing element in the partial matrix, return its local impact.
    """
    prng = check_random_state(random_state)
    global_impacts = {}
    edges = grone_order(partial)
    completed_partial = partial.copy()
    maxdet_complete_glb = maxdet_complete(completed_partial.copy())
    sgn, log_maxdet_global = np.linalg.slogdet(maxdet_complete_glb)
    for e_i, e in enumerate(edges):
        i, j = e
        local_partial = get_largest_submatrix(completed_partial, i, j)
        center, interval = feasible_interval(local_partial)

        total_impact = 0
        for retry in range(retries):
            attempts = 1
            rand_fill = prng.uniform(interval[0], interval[1])
            completed_partial[i, j] = completed_partial[j, i] = rand_fill
            randdet_complete_glb = maxdet_complete(completed_partial.copy())
            sgn, log_randdet_global = np.linalg.slogdet(randdet_complete_glb)
            while sgn < 0 and attempts < max_attempts:
                rand_fill = prng.uniform(interval[0], interval[1])
                completed_partial[i, j] = completed_partial[j, i] = rand_fill
                randdet_complete_glb = maxdet_complete(completed_partial.copy())
                sgn, log_randdet_global = np.linalg.slogdet(randdet_complete_glb)
                max_attempts += 1
            if sgn < 0:
                raise ValueError('Choice for (%d, %d) results in negative determinant' % (i, j))

            impact = log_subtract(log_maxdet_global, log_randdet_global) -\
                log_maxdet_global
            total_impact += np.exp(impact)
        global_impacts[(i, j)] = total_impact / retries
        completed_partial[i, j] = completed_partial[j, i] = center
    return global_impacts


def _sampling_local_impact(pos, partial, retries=10, max_attempts=5, random_state=None,
                          weighted=False):
    """DEPRECATED"""
    i, j = pos
    prng = check_random_state(random_state)
    local_partial = get_largest_submatrix(partial, i, j)
    center, interval = feasible_interval(local_partial)

    if weighted:
        in_interval = local_partial[local_partial > interval[0]]
        in_interval = in_interval[in_interval < interval[1]]
        if len(in_interval):
            can_sample = True
        else:
            can_sample = False

    local_partial[0, -1] = local_partial[-1, 0] = center
    sgn, log_maxdet_local = np.linalg.slogdet(local_partial)
    total_impact = 0
    for retry in range(retries):
        attempts = 1
        rand_fill = prng.uniform(interval[0], interval[1])
        local_partial[0, -1] = local_partial[-1, 0] = rand_fill
        sgn, log_randdet_local = np.linalg.slogdet(local_partial)
        while sgn < 0 and attempts < max_attempts:
            if weighted and can_sample:
                rand_fill = in_interval[prng.randint(len(in_interval))]
            else:
                rand_fill = prng.uniform(interval[0], interval[1])
            local_partial[0, -1] = local_partial[-1, 0] = rand_fill
            sgn, log_randdet_local = np.linalg.slogdet(local_partial)
            max_attempts += 1
        if sgn < 0:
            raise ValueError('Choice for (%d, %d) results in negative determinant' % (i, j))
        if log_randdet_local > log_maxdet_local + 1e-6:
            print 'oops'
        elif log_randdet_local < log_maxdet_local + 1e-6 and \
                log_randdet_local > log_maxdet_local:
            impact = 0
        else:
            impact = log_subtract(log_maxdet_local, log_randdet_local) -\
                log_maxdet_local
        total_impact += np.exp(impact)
    local_partial[0, -1] = local_partial[-1, 0] = np.nan
    return total_impact / retries


def local_impact(pos, partial, use_data=False, normalize=True,
                 divide_by_detA=True):
    i, j = pos
    local_partial = get_largest_submatrix(partial, i, j)
    center, interval = feasible_interval(local_partial)
    valid_data = local_partial[~np.isnan(local_partial)]

    if use_data:
        in_interval = valid_data[valid_data > interval[0]]
        in_interval = in_interval[in_interval < interval[1]]
        if len(in_interval):
            can_sample = True
        else:
            can_sample = False

    if use_data and can_sample:
        randdet_local = function_det(in_interval, local_partial).mean()
    else:
        randdet_local = integral_function_det(interval[0],
                                              interval[1], local_partial)
        randdet_local /= (interval[1] - interval[0])
    n = local_partial.shape[0]
    if n > 2 and divide_by_detA:
        A = local_partial[1:n - 1, 1:n - 1]
        signA, logdetA = np.linalg.slogdet(A)
        randdet_local /= np.exp(logdetA)

    if normalize:
        local_partial[0, -1] = local_partial[-1, 0] = center
        sgn, log_maxdet_local = np.linalg.slogdet(local_partial)
        local_partial[0, -1] = local_partial[-1, 0] = np.nan

        impact = np.log(randdet_local) - log_maxdet_local
        return np.exp(impact)
    else:
        return randdet_local


def pre_local_impact(partial, use_data=False, normalize=True):
    local_impacts = {}
    edges = grone_order(partial)
    completed_partial = partial.copy()
    for e in edges:
        i, j = e
        local_partial, keep = get_largest_submatrix(completed_partial, i, j, return_indices=True)
        center, interval = feasible_interval(local_partial)
        local_partial_original = partial[keep, :][:, keep]
        valid_data = local_partial_original[~np.isnan(local_partial_original)]

        if use_data:
            can_sample = False
            in_interval = valid_data[valid_data > interval[0]]
            in_interval = in_interval[in_interval < interval[1]]
            if len(in_interval):
                # Sometimes we sample only the completed values. In that
                # case, the `valid_data` is a list of 0's, which is the same
                # as the center. The impact will be set to 0 below.
                can_sample = True

        if use_data and can_sample:
            randdet_local = function_det(in_interval, local_partial).mean()
        else:
            randdet_local = integral_function_det(interval[0],
                                                  interval[1], local_partial)
            randdet_local /= (interval[1] - interval[0])

        if normalize:
            local_partial[0, -1] = local_partial[-1, 0] = center
            sgn, log_maxdet_local = np.linalg.slogdet(local_partial)
            local_partial[0, -1] = local_partial[-1, 0] = np.nan

            log_randdet = np.log(randdet_local)
            maxdet = np.exp(log_maxdet_local)
            if log_maxdet_local <= np.log(randdet_local):
                # either airthmetic error or impact is 0
                impact = 0
            else:
                impact = log_subtract(log_maxdet_local, np.log(randdet_local)) -\
                    log_maxdet_local
                # impact = np.log(randdet_local) - log_maxdet_local
                #impact = np.exp(impact)
                impact = (np.exp(log_maxdet_local) + randdet_local) / np.exp(log_maxdet_local)
                # impact = (maxdet + randdet_local) / maxdet
                #impact = np.log(np.exp(- log_randdet + log_maxdet_local) + 1) + log_randdet - log_maxdet_local
                impact = np.log(randdet_local) - log_maxdet_local
                impact = np.exp(impact)

        else:
            impact = randdet_local
        local_impacts[(i, j)] = impact
        completed_partial[i, j] = completed_partial[j, i] = center
    return local_impacts


def smart_query(partial, mtx, budget, clique_tree=None, random_state=None,
                strategy='local',
                preglobal_impacts=None, prelocal_impacts=None,
                use_data=True, normalize=True, candidate_nodes=None,
                candidate_edges_bound=None):
    prng = check_random_state(random_state)
    random.seed(prng.randint(MAXINT))
    pre_impacts = None
    if clique_tree is None:
        clqtree = CliqueTree()
        clqtree.from_graph(make_mask_graph(partial))
    else:
        clqtree = clique_tree.copy()
    if candidate_nodes is not None and type(candidate_nodes) is int:
        orig_candidate_nodes = candidate_nodes
        candidate_nodes = candidate_nodes * partial.shape[0] // 100
        if candidate_nodes == 0:
            raise ValueError('Sample node percentage is too low: %d' %
                             orig_candidate_nodes)
    if candidate_edges_bound is not None:
        orig_candidate_edges_bound = candidate_edges_bound
        candidate_edges_bound = candidate_edges_bound *\
            partial.shape[0] ** 2 // 100 / 2
        if candidate_edges_bound == 0:
            raise ValueError('Sample edge percentage is too low: %d' %
                             orig_candidate_edges_bound)

    # if candidate_edges_bound == 1 and len(clqtree.insertable) > 1:
        # pool = [x for x in clqtree.insertable]
        # rand_edge = pool[prng.choice(len(pool), 1)]
        # clqtree.insertable = set([rand_edge])

    if strategy.startswith('prelocal'):
        if prelocal_impacts is None:
            pre_impacts = pre_local_impact(partial, use_data=use_data,
                                           normalize=normalize)
        else:
            pre_impacts = prelocal_impacts
    elif strategy.startswith('preglobal'):
        if preglobal_impacts is None:
            pre_impacts = pre_global_impact(partial, random_state=prng)
        else:
            pre_impacts = preglobal_impacts

    while budget > 0 and len(clqtree.insertable):
        best_option = (-float('inf'), None)
        edges = []
        values = []
        edges_to_check = clqtree.insertable
        for e in edges_to_check:
            i, j = e
            if strategy.startswith('local'):
                if strategy.endswith('ahead'):
                    temptree = clqtree.copy()
                    temptree.add_edge(*e)
                    partial[i, j] = partial[j, i] = feasible_interval(partial)[0]
                    impact = sum([local_impact(e2, partial, use_data=use_data,
                                               normalize=normalize)
                                  for e2 in temptree.insertable])
                    partial[i, j] = partial[j, i] = np.nan
                else:
                    impact = local_impact(e, partial, use_data=use_data,
                                          normalize=normalize)
                # print strategy, 'query:', (i, j), 'w/ impact', impact
            elif strategy is 'global':
                impact = global_impact(e, partial, random_state=prng)
            elif strategy.startswith('pre'):
                if strategy.endswith('ahead'):
                    temptree = clqtree.copy()
                    temptree.add_edge(*e)
                    impact = sum([pre_impacts[e2]
                                  for e2 in temptree.insertable])
                else:
                    impact = pre_impacts[e]
            if 'size' in strategy:
                submatrix = get_largest_submatrix(partial, e[0], e[1])
                size = submatrix.shape[0] * submatrix.shape[1]
                normalized_size = size / partial.shape[0] ** 2
                impact = normalized_size * impact
            if 'stoch' in strategy:
                edges.append(e)
                values.append(impact)
            else:
                if impact >= best_option[0]:
                    best_option = (impact, e)
        if 'stoch' in strategy:
            total_value = sum(values)
            if total_value > 0:
                probabilities = [v / total_value for v in values]
            else:
                probabilities = None
            best_edge_index = prng.choice(range(len(edges)), size=1,
                                          p=probabilities)[0]
            best_edge = edges[best_edge_index]
            best_impact = values[best_edge_index]
        else:
            best_impact, best_edge = best_option
        i, j = best_edge

        if candidate_nodes is None and candidate_edges_bound is None:
            clqtree.add_edge(i, j)
        else:
            clqtree.add_edge(i, j, update_insertable=False)
            clqtree.insertable = set()
            used_nodes = set()
            while not clqtree.insertable:
                sampled_nodes = set(clqtree.G.nodes())
                if candidate_nodes is not None:
                    if type(candidate_nodes) is int:
                        pool = sampled_nodes - used_nodes
                        sampled_nodes = prng.choice(list(pool), candidate_nodes)
                        used_nodes = used_nodes.union(sampled_nodes)
                    elif type(candidate_nodes) is list:
                        sampled_nodes = candidate_nodes
                    else:
                        raise TypeError('Candidate nodes should either be a list or a number.')

                for v in sampled_nodes:
                    clqtree.update_insertable(v, stop_at=candidate_edges_bound)
            # if candidate_edges_bound == 1 and len(clqtree.insertable) > 1:
                # clqtree.insertable = set([list(clqtree.insertable)[0]])
            if not clqtree.insertable:
                warn('No insertable edges in candidates. Expanding search.')
                for v in clqtree.G:
                    clqtree.update_insertable(v)

        # print strategy, 'picked query:', best_edge, 'w/ impact', best_impact
        partial[i, j] = partial[j, i] = _oracle(mtx, i, j)

        budget -= 1
    return partial, budget * 2


def complete_with_query(partial, mtx, budget, strategy='rnd',
                        clique_tree=None, seed=None, prelocal_impacts=None,
                        preglobal_impacts=None, plot_queries=False,
                        return_queried=False, candidate_nodes=None,
                        candidate_edges_bound=None):
    missing = count_nans(partial)
    if budget >= missing:
        active_complete = mtx.copy()
        budget_left = 0
    else:
        budget = budget // 2
        if 'data' in strategy:
            use_data = True
        else:
            use_data = False
        if 'norm' in strategy:
            normalize = True
        else:
            normalize = False
        if strategy is 'rnd':
            active_partial, budget_left = rand_query(partial=partial.copy(),
                                                     mtx=mtx,
                                                     budget=budget,
                                                     clique_tree=clique_tree,
                                                     random_state=seed)
        elif strategy is 'structure':
            active_partial, budget_left = greedy_structure(partial=partial.copy(),
                                                           mtx=mtx,
                                                           budget=budget,
                                                           clique_tree=clique_tree)
        elif 'local' in strategy:
            active_partial, budget_left = smart_query(partial=partial.copy(),
                                                      mtx=mtx,
                                                      budget=budget,
                                                      clique_tree=clique_tree,
                                                      random_state=seed,
                                                      strategy=strategy,
                                                      prelocal_impacts=prelocal_impacts,
                                                      use_data=use_data,
                                                      normalize=normalize,
                                                      candidate_nodes=candidate_nodes,
                                                      candidate_edges_bound=candidate_edges_bound)
            if type(budget_left) is tuple:
                print 'Negative log difference'
                return active_partial, budget_left
        elif 'global' in strategy:
            active_partial, budget_left = smart_query(partial=partial.copy(),
                                                      mtx=mtx,
                                                      budget=budget,
                                                      clique_tree=clique_tree,
                                                      random_state=seed,
                                                      strategy=strategy,
                                                      preglobal_impacts=preglobal_impacts,
                                                      use_data=use_data,
                                                      normalize=normalize)
        elif 'grone' in strategy:
            if 'rnd' in strategy:
                sample = True
            else:
                sample = False
            active_partial, budget_left = grone_query(partial=partial.copy(),
                                                      mtx=mtx,
                                                      budget=budget,
                                                      sample=sample)
        partial_queried = None
        if return_queried:
            partial_queried = active_partial.copy()
        if plot_queries:
            plot_boolean_matrix(active_partial)
        active_complete = maxdet_complete(active_partial)
        if return_queried:
            return active_complete, partial_queried
    return active_complete, None
