from sklearn.datasets import make_spd_matrix, make_sparse_spd_matrix
from sklearn.datasets import make_biclusters, make_blobs, make_checkerboard
from sklearn.utils import check_random_state
from numpy import ones, diag, fill_diagonal, nan, isnan, eye, amin, zeros
from numpy.linalg import det, slogdet
from numpy.random import RandomState
from scipy.sparse.linalg import svds
import cliquetree
import utils
from utils import count_nans
from graph_algorithms import make_mask_graph
import gurobipy as gurobi


def partial_diagonal(M, reveal_pcnt=None, diagonals=None):
    """Hide part of the matrix.

    Return a partial version of the given matrix, together with the cliquetree
    of that matrix.
    """
    n = M.shape[0]
    if reveal_pcnt is not None:
        reveal = reveal_pcnt * n ** 2 / 100
        diagonals = (reveal - n) / (2 * n - 4)
        diagonals = int(round(diagonals))
    elif diagonals is None:
        raise ValueError('Either the percentage of revealed elements or the number of diagonals is required.')

    mask = ones(M.shape)
    M_partial = diag(diag(M))
    fill_diagonal(mask, 0)
    for i in range(1, diagonals + 1):
        M_partial += diag(diag(M, i), i) + diag(diag(M, -i), -i)
        mask -= diag(diag(mask, i), i) + diag(diag(mask, -i), -i)
    M_partial[mask > 0] = nan
    c = cliquetree.CliqueTree()
    c.from_graph(make_mask_graph(M_partial))
    return (M_partial, c)


def generate_matrix(n_dim, random_state):
    """Generate a random positive definite matrix of a given dimension.
    """
    B = make_spd_matrix(n_dim=n_dim, random_state=random_state)
    return B.T.dot(B)


def to_lowrank(M, rank):
    u, s, v = svds(M, rank)
    return u.dot(diag(s)).dot(v)


def to_symmetric(M):
    return M.dot(M.T)


def make_sparse_spd(n, alpha, rank, random_state=None):
    """Generates a sparse, low-rank SPD matrix.
    """
    mtx = make_sparse_spd_matrix(n, alpha, random_state=random_state)
    mtx = generate_matrix(n, random_state)
    mtx += amin(mtx.diagonal()) * eye(n)
    mtx = to_lowrank(mtx, rank)
    return mtx


def hide_elements(mtx, k, permute_matrix=True, max_tries=10,
                  random_state=None):
    """Replaces k elements of the given matrix with numpy.nan.

    The process generates a partial matrix that can be reconstructed with the
    max-determinant algorithm. This basically means that the mask graph of
    the partial matrix is chordal.

    Parameters
    ----------
    mtx : matrix

    k : int
        The number of elements in the matrix to hide.

    permute_matrix : bool, default is True
        If True, the rows and columns in the original matrix will be permuted.

    max_tries : int
        Number of times to try repeating the generation if the resulting matrix
        does not have enough missing elements

    seed : int
        The seed for the pseudorandom-number generator.


    Returns
    -------
    mtx : matrix
        A copy of the original matrix with k elements hidden.
    """
    best_mtx = (None, mtx.shape[0] ** 2)
    for _ in range(max_tries):
        partial = _hide_elements(mtx, k, permute_matrix, random_state)
        hidden =isnan(partial).sum()
        if hidden <= k:
            return (partial,
                    cliquetree.CliqueTree()
                              .from_graph(make_mask_graph(partial)))
        if hidden < best_mtx[1]:
            best_mtx = (partial, hidden)
    return (best_mtx[0],
            cliquetree.CliqueTree().from_graph(make_mask_graph(best_mtx[0])))


def _hide_blocks(mtx, n_blocks, random_state=None):
    mask, rows, cols = make_biclusters(mtx.shape, n_clusters=n_blocks,
                                       noise=0., shuffle=False,
                                       random_state=random_state)
    mask = mask.dot(mask.T)
    partial = mtx.copy()
    partial[mask == 0] = nan
    return (partial,
            cliquetree.CliqueTree().from_graph(make_mask_graph(partial)))


def hide_blocks(mtx, n_blocks, reveal_pcnt, random_state=None,
                upper_bound=False):
    prng = check_random_state(random_state)
    m = gurobi.Model('biclusters')
    n = mtx.shape[0]
    reveal = reveal_pcnt * n ** 2 / 100

    x = {}
    for i in range(n_blocks):
        if upper_bound:
            ub = n / 5
        else:
            ub = n
        x[i] = m.addVar(lb=2, ub=ub, name='cluster' + str(i),
                        vtype=gurobi.GRB.INTEGER)
    totErr = m.addVar(name='totErr', lb=0)
    m.update()

    m.setObjective(totErr, gurobi.GRB.MINIMIZE)
    m.addConstr(totErr >= 0)
    m.addConstr(totErr >= gurobi.quicksum(x[i] * x[i]
                                          for i in range(n_blocks)) - reveal)
    m.addConstr(gurobi.quicksum(x[i] for i in range(n_blocks)) == n)
    m.setParam('OutputFlag', False)
    m.optimize()

    sizes = []
    for v in m.getVars():
        if 'cluster' in v.varName:
            sizes.append(int(round(v.x)))

    sizes = [sizes[i] for i in prng.permutation(len(sizes))]
    print 'Block sizes:', sizes
    mask = zeros(mtx.shape)
    prev_ix = 0
    for size in sizes:
        mask[prev_ix:prev_ix + size, prev_ix:prev_ix + size] = 1
        prev_ix += size

    partial = mtx.copy()
    partial[mask == 0] = nan
    c = cliquetree.CliqueTree()
    c.from_graph(make_mask_graph(partial))
    return (partial, c)


def _hide_elements(mtx, k, permute_matrix=True, random_state=None):
    """Helper function to hide element that does all the work.

    Parameters
    ----------
    mtx : matrix

    k : int
        The number of elements in the matrix to hide.

    permute_matrix : bool, default is True
        If True, the rows and columns in the original matrix will be permuted.

    seed : int
        The seed for the pseudorandom-number generator.


    Returns
    -------
    mtx : matrix
        A copy of the original matrix with k elements hidden.
    """
    prng = check_random_state(random_state)
    n = mtx.shape[0]
    if k > n ** 2 - n:
        raise ValueError('Value for k is too big: Need to keep the diagonal.')
    indices = list(range(n))
    partial = mtx.copy()
    if permute_matrix:
        prng.shuffle(indices)
        partial, inverse = utils.rearrange(partial, indices,
                                           return_inverse=True)

    c = cliquetree.CliqueTree()
    mask_graph = make_mask_graph(partial)
    c.from_graph(mask_graph)
    hidden = 0
    edges_to_hide = set()
    while hidden < k:
        c.update_deletable()
        candidates = list(c.deletable)
        if len(candidates) == 0:
            break
        prng.shuffle(candidates)
        random_edge = candidates[0]
        c.remove_edge(*random_edge)
        edges_to_hide.add(random_edge)
        hidden += 2

    # Now, hide all the non-picked elements
    for i, j in edges_to_hide:
        partial[i, j] = partial[j, i] = nan
    return partial


def reveal_elements(partial, mtx, k, max_tries=10, seed=None):
    hidden_num = count_nans(partial)
    print 'starting with', hidden_num

    best_mtx = (None, hidden_num)
    for _ in range(max_tries):
        partial_new = _reveal_elements(partial.copy(), mtx, k)
        hidden = count_nans(partial_new)
        print 'after addition', hidden
        if hidden <= hidden_num - k:
            return (partial_new,
                    cliquetree.CliqueTree()
                              .from_graph(make_mask_graph(partial_new)))
        print hidden
        if hidden > best_mtx[1]:
            best_mtx = (partial_new, hidden)
    return (best_mtx[0],
            cliquetree.CliqueTree().from_graph(make_mask_graph(best_mtx[0])))


def _reveal_elements(partial, mtx, k, random_state=None):
    prng = check_random_state(random_state)
    c = cliquetree.CliqueTree()
    c.from_graph(make_mask_graph(partial))
    revealed = 0
    edges_to_reveal = set()
    while revealed < k:
        candidates = list(c.insertable)
        if len(candidates) == 0:
            break
        prng.shuffle(candidates)
        random_edge = candidates[0]
        c.add_edge(*random_edge)
        edges_to_reveal.add(random_edge)
        revealed += 2

    # Now, reveal all the selected elements
    for i, j in edges_to_reveal:
        partial[i, j] = partial[j, i] = mtx[i, j]
    return partial


def _deprecated_hide_elements(mtx, k, permute_matrix=True, seed=None):
    """Helper function to hide element that does all the work.

    Parameters
    ----------
    mtx : matrix

    k : int
        The number of elements in the matrix to hide.

    permute_matrix : bool, default is True
        If True, the rows and columns in the original matrix will be permuted.

    seed : int
        The seed for the pseudorandom-number generator.


    Returns
    -------
    mtx : matrix
        A copy of the original matrix with k elements hidden.
    """
    prng = RandomState(seed)
    n = mtx.shape[0]
    if k > n ** 2 - n:
        raise ValueError('Value for k is too big: Need to keep the diagonal.')
    indices = list(range(n))
    partial = mtx.copy()
    if permute_matrix:
        prng.shuffle(indices)
        partial, inverse = utils.rearrange(partial, indices,
                                           return_inverse=True)

    c = cliquetree.CliqueTree()
    core_edges = [(i, i + 1) for i in range(n - 1)]
    prng.shuffle(core_edges)
    added = 0
    elements_to_keep = n ** 2 - k - n
    picked = set()
    while added < elements_to_keep:
        if not c.insertable:
            if core_edges:
                random_edge = core_edges.pop()
            else:
                break
        else:
            candidates = [e for e in c.insertable]
            random_index = prng.randint(len(candidates))
            random_edge = candidates[random_index]
            # should not count the diagonal elements, only the off-diagonal
        added += 2
        picked.add(random_edge)
        c.add_edge(*random_edge)

    # Now, hide all the non-picked elements
    for i in range(n):
        for j in range(n):
            if i >= j:
                continue
            if (i, j) not in picked:
                partial[i][j] = partial[j][i] = nan
    return partial


def make_aTa(n, n_clusters, noise=0.5, minval=0.1, maxval=10, seed=None):
    mtx, rows, cols = make_biclusters((n, n), n_clusters=n_clusters,
                                      noise=noise,
                                      shuffle=False,
                                      minval=minval,
                                      maxval=maxval,
                                      random_state=seed)
    prng = RandomState(seed)
    mtx += + prng.rand(*mtx.shape)
    mtx = mtx.dot(mtx.T)
    assert utils.is_pd(mtx)
    assert utils.is_symmetric(mtx)
    assert det(mtx) > 0
    return mtx


def make_checkers(n, n_features=10, n_clusters=4, box_radius=6, cluster_std=.5,
                  random_state=None):
    positive_definite = False
    n_samples = n
    while not positive_definite:
        prng = check_random_state(random_state)

        centers = prng.rand(n_clusters, n_features) * box_radius
        samples_per_cluster = []
        splits = sorted(prng.choice(range(n_samples - 1), n_clusters - 1,
                                    replace=False))
        splits = [s + 1 for s in splits]
        samples_per_cluster.append(splits[0])
        for i in range(1, n_clusters - 1):
            # r = prng.randint(1, available_samples - (n_clusters - i))
            size = splits[i] - splits[i - 1]
            samples_per_cluster.append(size)
        samples_per_cluster.append(n_samples - splits[-1])

        data = zeros((n_samples, n_features))
        up_to = 0
        for i in range(n_clusters):
            data_cl, labels_cl = make_blobs(n_samples=samples_per_cluster[i],
                                            n_features=n_features,
                                            centers=centers[i, :].reshape(1, -1),
                                            cluster_std=cluster_std,
                                            shuffle=False,
                                            random_state=random_state)
            data[up_to: up_to + samples_per_cluster[i], :] = data_cl
            up_to += samples_per_cluster[i]
        data += prng.rand(*data.shape)
        data = data.dot(data.T)
        data += prng.rand(*data.shape)
        data = data.dot(data.T)

        sgn, logdet = slogdet(data)
        if sgn > 0 and logdet > 0 and utils.is_pd(data):
            positive_definite = True
    return data


def make_checkers2(n, n_clust, noise=0.3, minval=1, maxval=10,
                   random_state=None):
    A, rows, cols = make_checkerboard((n, n), n_clusters=n_clust, noise=noise,
                                      minval=minval, maxval=maxval,
                                      shuffle=False, random_state=random_state)
    return A
