import functools
import os.path
import tempfile
import time
from collections import defaultdict

import algorithms
import data_generator as data_gen
import dill as pickle
import ipyparallel as ipp
import numpy as np
import pandas as pd
import scipy
import solvers
import utils
import seaborn as sns
from matplotlib import pyplot as plt
from numpy.random import RandomState

result_dir = 'result_files'
result_df_filename = 'results.csv'
result_hidden_df_filename = 'hidden_results.csv'
columns = ['data_seed', 'hiding_seed', 'seed', 'hiding', 'reveal_pcnt',
           'method', 'true_logdet', 'max_logdet', 'result_file'
           ]
hidden_columns = ['data_seed', 'hiding_seed', 'seed', 'hiding', 'budget',
                  'method', 'true_logdet', 'max_logdet', 'result_file'
                  ]

def profile_run(partial, mtx, budgets, strategy='rnd'):
    true_logdet = np.linalg.slogdet(mtx)[1]
    candidate_nodes = None
    candidate_edges_bound = None
    if '@' in strategy:
        opts = strategy.split('@')[1]
        splits = opts.split('_')
        candidate_nodes = int(splits[0])
        if len(splits) > 1:
                candidate_edges_bound = int(splits[1])

    prev_budget = 0
    partial_queried = partial.copy()
    for b_i, budget in enumerate(budgets):
        tic = time.time()
        extra_budget = budget - prev_budget
        print budget, extra_budget, utils.count_nans(partial_queried),
        active_complete, partial_queried =\
            algorithms.complete_with_query(partial_queried,
                                           mtx,
                                           extra_budget,
                                           strategy=strategy,
                                           seed=None,
                                           return_queried=True,
                                           candidate_nodes=candidate_nodes,
                                           candidate_edges_bound=candidate_edges_bound)
        #if partial_queried is not None:
        #    utils.plot_boolean_matrix(partial_queried)
        sign, logdet = np.linalg.slogdet(active_complete)
        prev_budget = budget
        print logdet, true_logdet, time.time() - tic


def run_random_query(matrix_info, budgets):
    partial, mtx, true_logdet = matrix_info
    rand_performance = [0 for i in range(len(budgets))]
    times = [0 for i in range(len(budgets))]
    prev_budget = 0
    partial_queried = partial.copy()
    tic = time.time()
    for b_i, budget in enumerate(budgets):
        extra_budget = budget - prev_budget
        active_complete, partial_queried =\
            algorithms.complete_with_query(partial_queried,
                                           mtx,
                                           extra_budget,
                                           strategy='rnd',
                                           seed=None,
                                           return_queried=True)
        sign, logdet = np.linalg.slogdet(active_complete)
        rand_performance[b_i] = logdet
        times[b_i] = time.time() - tic
        prev_budget = budget
    return rand_performance, times


def run_smart_query(call_args, budgets, plot_partial=False):
    partial, mtx, true_logdet, strategy = call_args
    performance = [0 for i in range(len(budgets))]
    times = [0 for i in range(len(budgets))]
    prev_budget = 0
    partial_queried = partial.copy()
    pre_local_imp = None
    if 'data' in strategy:
        use_data = True
    else:
        use_data = False
    if 'norm' in strategy:
        normalize = True
    else:
        normalize = False
    if strategy.startswith('pre'):
        pre_local_imp = algorithms.pre_local_impact(partial, use_data=use_data,
                                                    normalize=normalize)
    candidate_nodes = None
    candidate_edges_bound = None
    if '@' in strategy:
        opts = strategy.split('@')[1]
        splits = opts.split('_')
        candidate_nodes = int(splits[0])
        if len(splits) > 1:
                candidate_edges_bound = int(splits[1])

    tic = time.time()
    for b_i, budget in enumerate(budgets):
        extra_budget = budget - prev_budget
        print strategy
        active_complete, partial_queried = \
            algorithms.complete_with_query(partial_queried,
                                           mtx,
                                           extra_budget,
                                           strategy=strategy,
                                           seed=None,
                                           return_queried=True,
                                           prelocal_impacts=pre_local_imp,
                                           candidate_nodes=candidate_nodes,
                                           candidate_edges_bound=candidate_edges_bound)
        times[b_i] = time.time() - tic
        if partial_queried is not None and plot_partial:
            utils.plot_boolean_matrix(partial_queried)
        sign, logdet = np.linalg.slogdet(active_complete)
        performance[b_i] = logdet  # - true_logdet
        prev_budget = budget
    return performance, times


def generate_mtx(n, n_clusters=5, n_features=5, box_radius=2, data_seed=None):
    if data_seed is None:
        data_seed = np.random.randint(utils.MAXINT)
    mtx = data_gen.make_checkers(n=n, n_features=n_features,
                                 n_clusters=n_clusters,
                                 box_radius=box_radius,
                                 random_state=data_seed)
    return mtx


def budget_exp(strategies, percentages, n=None, workers=None, hiding='rnd',
               num_repeats=1, num_experiments=1, reveal_pcnt=25, n_blocks=5,
               seed=None, real_dataset=None):
    """Runs a budget experiment.

    n : int, default is 50
        The size of the matrix
    num_repeats : int, default is 1
        The number of times that a method is going to be executed. The result
        is the aggregate (average) of all the runs.
    num_experiments : int, default is 1
        The number of experiments to run.
    reveal_pcnt : int, default is 25
        The percentage of entries in the matrix that will be revealed.
    """
    if workers is None:
        rc = ipp.Client()
        workers = rc.load_balanced_view()
    # If the filenames below change, also update the `write_results` function
    if not os.path.isfile(result_df_filename):
        result_df = pd.DataFrame(columns=columns)
        result_df.to_csv(result_df_filename, header=True, index=False)
    else:
        result_df = pd.read_csv(result_df_filename)
    if not os.path.exists(result_dir):
        os.makedirs(result_dir)

    averages = {}
    mtx_stats = []

    tic = time.time()
    if seed is None:
        seed = np.random.randint(utils.MAXINT)
    prng = RandomState(seed)

    for _round in range(num_experiments):
        if real_dataset is not None:
            splits = real_dataset.split('_')
            dataset = splits[0]
            covariance = False
            if len(splits) > 1:
                covariance = True
            mtx = load_real_dataset(dataset, return_covariance=covariance, size=n)
            n = mtx.shape[0]
            data_seed = real_dataset
        else:
            data_seed = prng.randint(utils.MAXINT)
            mtx = generate_mtx(n=n, data_seed=data_seed)
        hiding_seed = prng.randint(utils.MAXINT)
        print 'Seeds:', data_seed, hiding_seed

        if hiding == 'rnd':
            revealed_elems = (n ** 2) * reveal_pcnt // 100
            if not revealed_elems % 2 == 0:
                revealed_elems += 1
            hide_elements = (n ** 2) - revealed_elems
            partial, c = data_gen.hide_elements(mtx, k=hide_elements,
                                                permute_matrix=False,
                                                random_state=hiding_seed)
        elif hiding == 'blocks':
            partial, c = data_gen.hide_blocks(mtx, n_blocks=n_blocks,
                                              reveal_pcnt=reveal_pcnt,
                                              random_state=hiding_seed,
                                              upper_bound=False)
        elif hiding == 'diags':
            partial, c = data_gen.partial_diagonal(mtx,
                                                   reveal_pcnt=reveal_pcnt)
        else:
            raise ValueError('`hiding` should be one of rnd, blocks or diags.')

        hidden = utils.count_nans(partial)
        budgets = []
        budgets.extend([hidden * i // 100 for i in percentages])

        complete = solvers.maxdet_complete(partial.copy())
        complete_logdet = np.linalg.slogdet(complete)[1]
        true_logdet = np.linalg.slogdet(mtx)[1]
        print "MaxDet", complete_logdet
        print "Original logdet", true_logdet
        # TODO: Save complete_logdet and true_logdet to DF
        mtx_stats.append((complete_logdet, true_logdet))
        for b_i, budget in enumerate(budgets):
            percentage = percentages[b_i]
            if percentage not in averages:
                averages[percentage] = defaultdict(list)

        # Random goes first
        if 'rnd' in strategies:
            tic_strategy = time.time()
            result = workers.map_sync(functools.partial(run_random_query,
                                                        budgets=budgets),
                                      [(partial.copy(), mtx, true_logdet)
                                       for i in range(num_repeats)])
            rand_performance = []
            rand_time = []
            for i in range(len(result[0][0])):
                total_perf = 0
                total_time = 0
                for partial_run in result:
                    perf_i = partial_run[0][i]
                    time_i = partial_run[1][i]
                    total_perf += perf_i
                    total_time += time_i
                rand_performance.append(total_perf)
                rand_time.append(total_time)
            rand_performance = [x / len(result) for x in rand_performance]
            rand_time = [x / len(result) for x in rand_time]
            for b_i, budget in enumerate(budgets):
                percentage = percentages[b_i]
                averages[percentage]['rnd'].append((rand_performance[b_i],
                                                    rand_time[b_i]))
            print 'random done', '%.2f (%.2f)' % (time.time() - tic_strategy,
                                                  time.time() - tic)

        # Next are the sampled ones. Repeat them and average the result
        for strategy in [strat_i for strat_i in strategies if '@' in strat_i]:
            tic_strategy = time.time()
            result = workers.map_sync(functools.partial(run_smart_query,
                                                        budgets=budgets),
                                      [(partial.copy(), mtx,
                                        true_logdet, strategy)
                                       for i in range(num_repeats)])
            sampling_performance = []
            sampling_time = []
            for i in range(len(result[0][0])):
                total_perf = 0
                total_time = 0
                for partial_run in result:
                    perf_i = partial_run[0][i]
                    time_i = partial_run[1][i]
                    total_perf += perf_i
                    total_time += time_i
                sampling_performance.append(total_perf)
                sampling_time.append(total_time)
            sampling_performance = [x / len(result)
                                    for x in sampling_performance]
            sampling_time = [x / len(result) for x in sampling_time]
            for b_i, budget in enumerate(budgets):
                pcnt = percentages[b_i]
                averages[pcnt][strategy].append((sampling_performance[b_i],
                                                 sampling_time[b_i]))
            print strategy, 'done %.2f (%.2f)' % (time.time() - tic_strategy,
                                                  time.time() - tic)

        # Run strategies
        tic_strategy = time.time()
        determ_strategies = [strat_i
                             for strat_i in strategies
                             if '@' not in strat_i and 'rnd' not in strat_i]
        if determ_strategies:
            result = workers.map_sync(functools.partial(run_smart_query,
                                                        budgets=budgets),
                                      [(partial.copy(), mtx, true_logdet,
                                        strategy)
                                       for strategy in determ_strategies])
            for i, strategy in enumerate(determ_strategies):
                worker_result = result[i]
                for b_i, budget in enumerate(budgets):
                    performance_i = worker_result[0][b_i]
                    time_i = worker_result[1][b_i]
                    percentage = percentages[b_i]
                    averages[percentage][strategy].append((performance_i, time_i))
            print 'Strategies done', '%.2f (%.2f)' % (time.time() - tic_strategy,
                                                      time.time() - tic)

        args = dict()
        for col in columns:
            if col.startswith('local'):
                if col in strategies:
                    args[col] = True
                else:
                    args[col] = False
        args['hiding'] = hiding
        args['seed'] = seed
        args['data_seed'] = data_seed
        args['hiding_seed'] = hiding_seed
        args['reveal_pcnt'] = reveal_pcnt
        args['true_logdet'] = true_logdet
        args['max_logdet'] = complete_logdet
        result_df = write_results(args, result_df, strategies=strategies,
                                  averages=averages)


def hidden_exp(strategies, percentages, n=None, workers=None, hiding='rnd',
               num_repeats=1, num_experiments=1, budget=10, n_blocks=5,
               seed=None, real_dataset=None):
    """Runs an experiment for different number of hidden elements.

    n : int, default is 50
        The size of the matrix
    num_repeats : int, default is 1
        The number of times that a method is going to be executed. The result
        is the aggregate (average) of all the runs.
    num_experiments : int, default is 1
        The number of experiments to run.
    budget: int, default is 10
        The percentage of entries in the matrix that will be queried.
    """
    if workers is None:
        rc = ipp.Client()
        workers = rc.load_balanced_view()
    # If the filenames below change, also update the `write_results` function
    if not os.path.isfile(result_hidden_df_filename):
        result_df = pd.DataFrame(columns=hidden_columns)
        result_df.to_csv(result_hidden_df_filename, header=True, index=False)
    else:
        result_df = pd.read_csv(result_hidden_df_filename)
    if not os.path.exists(result_dir):
        os.makedirs(result_dir)

    averages = {}
    mtx_stats = []

    tic = time.time()
    if seed is None:
        seed = np.random.randint(utils.MAXINT)
    prng = RandomState(seed)

    for _round in range(num_experiments):
        if real_dataset is not None:
            splits = real_dataset.split('_')
            dataset = splits[0]
            covariance = False
            if len(splits) > 1:
                covariance = True
            mtx = load_real_dataset(dataset, return_covariance=covariance, size=n)
            n = mtx.shape[0]
            data_seed = real_dataset
        else:
            data_seed = prng.randint(utils.MAXINT)
            mtx = generate_mtx(n=n, data_seed=data_seed)
        hiding_seed = prng.randint(utils.MAXINT)
        print 'Seeds:', data_seed, hiding_seed

        for reveal_pcnt in percentages:
            if reveal_pcnt not in averages:
                averages[reveal_pcnt] = defaultdict(list)

        for reveal_pcnt in percentages:
            if hiding == 'rnd':
                revealed_elems = (n ** 2) * reveal_pcnt // 100
                if not revealed_elems % 2 == 0:
                    revealed_elems += 1
                hide_elements = (n ** 2) - revealed_elems
                partial, c = data_gen.hide_elements(mtx, k=hide_elements,
                                                    permute_matrix=False,
                                                    random_state=hiding_seed)
            elif hiding == 'blocks':
                if reveal_pcnt > 50:
                    upper_bound = False
                    print 'unbounded QIP'
                else:
                    upper_bound = True
                partial, c = data_gen.hide_blocks(mtx, n_blocks=n_blocks,
                                                  reveal_pcnt=reveal_pcnt,
                                                  random_state=hiding_seed,
                                                  upper_bound=False)
            elif hiding == 'diags':
                partial, c = data_gen.partial_diagonal(mtx,
                                                       reveal_pcnt=reveal_pcnt)
            else:
                raise ValueError('`hiding` should be one of rnd, blocks or diags.')

            hidden = utils.count_nans(partial)
            budgets = []
            budgets.extend([hidden * i // 100 for i in [budget]])

            complete = solvers.maxdet_complete(partial.copy())
            complete_logdet = np.linalg.slogdet(complete)[1]
            true_logdet = np.linalg.slogdet(mtx)[1]
            print "MaxDet", complete_logdet
            print "Original logdet", true_logdet
            # TODO: Save complete_logdet and true_logdet to DF
            mtx_stats.append((complete_logdet, true_logdet))

            # Random goes first
            if 'rnd' in strategies:
                tic_strategy = time.time()
                result = workers.map_sync(functools.partial(run_random_query,
                                                            budgets=budgets),
                                          [(partial.copy(), mtx, true_logdet)
                                           for i in range(num_repeats)])
                rand_performance = []
                rand_time = []
                for i in range(len(result[0][0])):
                    total_perf = 0
                    total_time = 0
                    for partial_run in result:
                        perf_i = partial_run[0][i]
                        time_i = partial_run[1][i]
                        total_perf += perf_i
                        total_time += time_i
                    rand_performance.append(total_perf)
                    rand_time.append(total_time)
                rand_performance = [x / len(result) for x in rand_performance]
                rand_time = [x / len(result) for x in rand_time]
                averages[reveal_pcnt]['rnd'].append((rand_performance[0],
                                                    rand_time[0]))
                print 'random done', '%.2f (%.2f)' % (time.time() - tic_strategy,
                                                      time.time() - tic)

            # Next are the sampled ones. Repeat them and average the result
            for strategy in [strat_i for strat_i in strategies if '@' in strat_i]:
                tic_strategy = time.time()
                result = workers.map_sync(functools.partial(run_smart_query,
                                                            budgets=budgets),
                                          [(partial.copy(), mtx,
                                            true_logdet, strategy)
                                           for i in range(num_repeats)])
                sampling_performance = []
                sampling_time = []
                for i in range(len(result[0][0])):
                    total_perf = 0
                    total_time = 0
                    for partial_run in result:
                        perf_i = partial_run[0][i]
                        time_i = partial_run[1][i]
                        total_perf += perf_i
                        total_time += time_i
                    sampling_performance.append(total_perf)
                    sampling_time.append(total_time)
                sampling_performance = [x / len(result)
                                        for x in sampling_performance]
                sampling_time = [x / len(result) for x in sampling_time]
                averages[reveal_pcnt][strategy]\
                    .append((sampling_performance[0], sampling_time[0]))
                print strategy, 'done %.2f (%.2f)' % (time.time() - tic_strategy,
                                                      time.time() - tic)

            # Run strategies
            tic_strategy = time.time()
            determ_strategies = [strat_i
                                 for strat_i in strategies
                                 if '@' not in strat_i and 'rnd' not in strat_i]
            if determ_strategies:
                result = workers.map_sync(functools.partial(run_smart_query,
                                                            budgets=budgets),
                                          [(partial.copy(), mtx, true_logdet,
                                            strategy)
                                           for strategy in determ_strategies])
                for i, strategy in enumerate(determ_strategies):
                    worker_result = result[i]
                    performance_i = worker_result[0][0]
                    time_i = worker_result[1][0]
                    averages[reveal_pcnt][strategy]\
                        .append((performance_i, time_i))
                print 'Strategies done', '%.2f (%.2f)' % (time.time() - tic_strategy,
                                                          time.time() - tic)

            args = dict()
            args['hiding'] = hiding
            args['seed'] = seed
            args['data_seed'] = data_seed
            args['hiding_seed'] = hiding_seed
            args['budget'] = budget
            args['true_logdet'] = true_logdet
            args['max_logdet'] = complete_logdet
            result_df = write_results(args, result_df, strategies=strategies,
                                      averages=averages,
                                      result_df_filename=result_hidden_df_filename)


def write_results(args, df, strategies=None, averages=None,
                  result_dir='result_files', result_df_filename='results.csv'):
    with tempfile.NamedTemporaryFile(mode='w+b', suffix='.dat',
                                     dir=result_dir, delete=False) as temp:
        pickle.dump(averages, temp)
        for strategy in strategies:
            row = args
            row['method'] = strategy
            row['result_file'] = temp.name
            df = df.append(row, ignore_index=True)
            last_row_index = df.index[-1]
            with open(result_df_filename, 'a') as f:
                to_write = pd.DataFrame(df.loc[last_row_index]).T
                to_write.to_csv(f, header=False, index=False)
    return df


def parse_df(filename=None):
    if filename is None:
        filename = result_df_filename

    result_df = pd.read_csv(filename)
    return result_df


def load_results(df, index):
    row = df.loc[index]
    with open(row.result_file, 'r') as f:
        averages = pickle.load(f)
        return averages, row


def performance(averages, strategies):
    perf = {}
    for b in averages:
        perf[b] = {strategy: [] for strategy in strategies}
    for b in sorted(averages):
        for repeat in range(len(averages[b]['rnd'])):
            base_perf, base_time = averages[b]['rnd'][repeat]
            #print base
            for strategy in strategies:
                if len(averages[b][strategy]) == 0:
                    continue
                strategy_perf, strategy_time = averages[b][strategy][repeat]
                perf_ratio = np.exp(strategy_perf - base_perf)
                perf[b][strategy].append((perf_ratio, strategy_time))

    return perf


def load_real_dataset(dataset, return_covariance=False, size=None):
    if dataset == 'protein':
        # For the protein dataset, keep the rows without zeros
        mtx = scipy.io.loadmat('protein.mat')['data']
        i, j = np.where(mtx == 0)
        keep_rows = [ii for ii in range(mtx.shape[0])
                     if ii not in set(i)]
        mtx = mtx[keep_rows, :]
    elif dataset in ['movielens', 'dating']:
        if dataset == 'movielens':
            mtx = scipy.io.loadmat('movielensSubM.mat')['smallM']
        elif dataset == 'dating':
            mtx = scipy.io.loadmat('dating.mat')['smallM']

        rows = np.zeros(mtx.shape[0])
        cols = np.zeros(mtx.shape[1])
        for i in range(mtx.shape[0]):
            rows[i] = mtx.getrow(i).data.mean()
        for i in range(mtx.shape[1]):
            cols[i] = mtx.getcol(i).data.mean()
        mtx = mtx.todense()
        i, j = np.where(mtx == 0)
        for ii, jj in zip(i, j):
                mtx[ii, jj] = (rows[ii] + cols[jj]) / 2
        if size is not None:
            mtx = mtx[:size, :size]
        mtx = np.array(mtx)
    else:
        raise ValueError('Dataset not available.')

    n, m = mtx.shape

    if n > m:
        thin = True
    else:
        thin = False
    if return_covariance:
        pd_mtx = mtx.T.dot(mtx)
        if not thin and n != m:
            pd_mtx += np.eye(m)
    else:
        pd_mtx = mtx.dot(mtx.T)
        if thin and n != m:
            pd_mtx += np.eye(n)
    return pd_mtx


def load_for_plot(result_df, plot_methods, plot_hiding, real_dataset=None):
    if real_dataset is not None:
        result_df = result_df[result_df.data_seed == real_dataset]
        grps = result_df.groupby(['hiding_seed', 'hiding'])
    else:
        grps = result_df.groupby(['data_seed', 'hiding'])
    all_avg = defaultdict(lambda: defaultdict(list))
    for name, grp_df in grps:
        if name[1] != plot_hiding:
            continue
        methods = grp_df.groupby(grp_df.method)
        if not all([x in methods.grouper.groups
                    for x in plot_methods + ['rnd']]):
            continue
        base_idx = max(methods.grouper.groups['rnd'])
        curr, row = load_results(result_df, base_idx)
        rnd_perf = {}
        local_perf = {}
        for pcnt in curr:
            rnd_perf[pcnt] = curr[pcnt]['rnd']
            local_perf[pcnt] = curr[pcnt]['local']

        for method, idx in methods.grouper.groups.iteritems():
            if max(idx) == base_idx:
                continue
            method_idx = max(idx)
            curr, row = load_results(result_df, method_idx)
            for pcnt in curr:
                logratio = curr[pcnt][method][0][0] - rnd_perf[pcnt][0][0]
                time_ratio = local_perf[pcnt][0][1] * 1. / curr[pcnt][method][0][1]
                #time_ratio = curr[pcnt][method][0][1]
                all_avg[method][pcnt].append((logratio, time_ratio))

    methods_det = {}
    methods_time = {}
    for method in plot_methods:
        method_det = [[all_avg[method][pcnt][exp][0]
                       for exp in range(len(all_avg[method][pcnt]))]
                      for pcnt in sorted(all_avg[method])
                      ]
        method_time = [[all_avg[method][pcnt][exp][1]
                        for exp in range(len(all_avg[method][pcnt]))]
                       for pcnt in sorted(all_avg[method])]
        methods_det[method] = method_det
        methods_time[method] = method_time
    return methods_det, methods_time


def plot_boxes(fig, data, plot_methods, percentages, width=0.4, ytickstep=1,
               ncol=3):
    # Create an axes instance
    sns.set_style('white')
    ax = fig.add_subplot(111)

    ## add patch_artist=True option to ax.boxplot()
    ## to get fill color
    num_methods = len(plot_methods)
    group_width = ((num_methods - 1) * (width / 2) + num_methods * width)
    positions = [i * group_width + i for i in range(len(percentages))]
    colors = [
        "#568335",
        "#7C45DC",
        "#73D74D",
        "#D2428D",
        "#CBD141",
        "#7576CF",
        "#BC822D",
        "#548EB1",
        "#D04D2B",
        "#53D29F",
        "#CA4CD1",
        "#C2505A",
        "#AE6DA9"
    ]

    boxes = []
    labels = []
    for i, method in enumerate(plot_methods):
        data_to_plot = data[method]
        for j, pos in enumerate(positions):
            bp = ax.boxplot(data_to_plot[j], patch_artist=True,
                            whis='range',
                            widths=width,
                            positions=[pos + 0.6 * i],
                            manage_xticks=False)
            if j == 0:
                boxes.append(bp['boxes'][0])
                if '@' not in method:
                    labels.append('Select')
                else:
                    second_part = method.split('@')[1]
                    splits = second_part.split('_')
                    if len(splits) > 1:
                        # sample nodes + edges
                        sample_nodes = splits[0]
                        sample_edges = splits[1]
                        labels.append('XSelect(%s, %s)' % (sample_nodes, sample_edges))
                    else:
                        sample_nodes - splits[0]
                        labels.append('XSelect(%s, _)' % (sample_nodes))
            for box in bp['boxes']:
                box.set(color='#3d3d3d', linewidth=2)
                box.set(facecolor=colors[i])

            for whisker in bp['whiskers']:
                whisker.set(color='#3d3d3d', linewidth=2)

            for cap in bp['caps']:
                cap.set(color='#3d3d3d', linewidth=2)

            for median in bp['medians']:
                median.set(color='#3d3d3d', linewidth=2)

    ax.set_xticks([p + group_width / 2 for p in positions])
    ax.set_xticklabels(percentages)
    _, max_x = ax.get_xlim()
    ax.set_xlim(-width, max_x)
    min_y, max_y = ax.get_ylim()
    y_gap = (max_y - min_y) / 50.
    ax.set_yticks(np.arange(min_y, max_y + ytickstep, ytickstep))
    ax.set_ylim(min_y - y_gap, max_y + y_gap)

    ax.tick_params('y', length=8, which='major', width=3, pad=15,
                   direction='out', right=False, left=True, labelsize=30)
    ax.tick_params('x', length=8, which='major', width=3, pad=5,
                   direction='out', top=False, bottom=True, labelsize=30)
    ax.spines['left'].set_linewidth(3)
    ax.spines['bottom'].set_linewidth(3)

    plt.legend(boxes, labels, bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
               mode='expand', ncol=ncol, borderaxespad=0.,
               frameon=False, prop={'size': 20})
    # plt.legend(boxes, labels, bbox_to_anchor=(1.05, 1), loc=2,
               # borderaxespad=0., frameon=False, prop={'size':20})

    sns.despine(top=True, right=True, bottom=False, offset=10, trim=True)

    sns.set_style(None)
    return fig


def plot_boxes_time(fig, data, plot_methods, percentages, ncol=2, ytickstep=1):
    # Create an axes instance
    sns.set_style('white')
    ax = fig.add_subplot(111)

    ## add patch_artist=True option to ax.boxplot()
    ## to get fill color
    colors = [
        "#568335",
        "#7C45DC",
        "#73D74D",
        "#D2428D",
        "#CBD141",
        "#7576CF",
        "#BC822D",
        "#548EB1",
        "#D04D2B",
        "#53D29F",
        "#CA4CD1",
        "#C2505A",
        "#AE6DA9"
    ]

    labels = []
    markers = ['square', 'o', '^', 'D', 's']
    for i, method in enumerate(plot_methods):
        if method == 'local':
            continue
        data_to_plot = data[method]
        points = [np.mean(data_to_plot[j])
                  for j, pcnt in enumerate(percentages)]
        print method
        print points
        print

        if '@' not in method:
            labels.append('Select')
        else:
            second_part = method.split('@')[1]
            splits = second_part.split('_')
            if len(splits) > 1:
                # sample nodes + edges
                sample_nodes = splits[0]
                sample_edges = splits[1]
                labels.append('XSelect(%s, %s)' % (sample_nodes, sample_edges))
            else:
                sample_nodes - splits[0]
                labels.append('XSelect(%s, _)' % (sample_nodes))
        label = labels[-1]
        print markers[i]
        markersize = 15
        if markers[i] == 's':
            markersize = 13
        ax.plot(range(len(points)), points, marker=markers[i], markersize=markersize,
                linewidth=3, color=colors[i], label=label)

    ax.set_xticks(range(len(percentages)))
    ax.set_xticklabels(percentages)
    min_x, max_x = ax.get_xlim()
    min_y, max_y = ax.get_ylim()
    x_gap = (max_x - min_x) / 50.
    y_gap = (max_y - min_y) / 50.
    ax.set_yticks(np.arange(min_y, max_y + ytickstep, ytickstep))

    ax.set_xlim(min_x - x_gap, max_x + x_gap)
    ax.set_ylim(min_y - y_gap, max_y + y_gap)

    ax.tick_params('y', length=8, which='major', width=3, pad=10,
                   direction='out', right=False, left=True, labelsize=30)
    ax.tick_params('x', length=8, which='major', width=3, pad=5,
                   direction='out', top=False, bottom=True, labelsize=30)
    ax.spines['left'].set_linewidth(3)
    ax.spines['bottom'].set_linewidth(3)

    plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
               mode='expand', ncol=ncol, borderaxespad=0.,
               frameon=False, prop={'size': 25})
    # plt.legend(bbox_to_anchor=(1.05, 1), loc=2, mode='expand',
    #            borderaxespad=0., frameon=False, prop={'size': 20})

    sns.despine(top=True, right=True, bottom=False, offset=15, trim=True)

    sns.set_style(None)
    return fig
