from __future__ import division

import numpy as np

from solvers import maxdet_complete
from data_generator import partial_diagonal, generate_matrix
from strategies import strategy_inner_order, strategy_outer_order
import matrix_operations as matop
import graph_algorithms as graphalg


def oracle(M, i, j):
    """Returns the true value of a missing element.
    """
    return M[i, j]


def complete(M_partial, mask, missing, oracle, budget=0, M_original=None,
             strategy=None, test_cell=None, test_fill='min', prng=None,
             intervals=None, local_impact=None):
    """
    Parameters
    ----------
    M_partial : numpy.matrix
        The matrix with the missing values.
    M_original : numpy.matrix
        The original matrix. Required for the oracle.
    mask : numpy.matrix
        A matrix, of the same size as M_partial, with ones at the unknown
        positions and zeros everywhere else.
    missing : float
        The number indicating that an element is missing.
    oracle : function
        A function that returns the true value of a cell
    budget : int
        The number of times the oracle can be asked.
    strategy : function
        The ranking of elements on which to call the oracle.
    """

    perfect_elim_order = graphalg.minDegreeAlgorithm(M_partial, missing)
    perfect_elim_order.reverse()
    inverting_order = [perfect_elim_order.index(x)
                       for x in range(M_partial.shape[0])]
    n = M_partial.shape[0]
    M_rearranged = M_partial[perfect_elim_order, :][:, perfect_elim_order]
    mask = mask[perfect_elim_order, :][:, perfect_elim_order]
    edges = ((j, i) for j in range(M_partial.shape[0])
             for i in range(j, -1, -1))
    # traverse and fill missing edges of M_partial according to the reversed
    # perfect elimination ordering
    budget_ranking = []
    if strategy is not None:
        budget_ranking = strategy(mask)
        rank_index = 0
        print 'budget allocation'
        while budget > 0 and rank_index < len(budget_ranking):
            (i, j) = budget_ranking[rank_index]
            if (i, j) == (0, n - 1) or (i, j) == (n - 1, 0):
                rank_index += 1
                continue
            print (i, j),
            M_rearranged[i, j] = M_rearranged[j, i] = oracle(M_original,
                                                             inverting_order[i],
                                                             inverting_order[j])
            budget -= 1
            rank_index += 1
        print
        print M_rearranged
        print

    for (j, i) in edges:
        if M_rearranged[i, j] != missing:
            #print 'Oracle det (%d, %d):  %.5g' % \
                #(i, j, np.linalg.det(matop.getSubMtx(M_rearranged, i, j)))
            continue
        M_submatrix = matop.getSubMtx(M_rearranged, i, j)
        (maxval, interval) = maxdet_complete(M_submatrix)
        if intervals is not None:
            intervals[(inverting_order[i], inverting_order[j])] = (interval[0], interval[1])
        M_rearranged[i, j] = M_rearranged[j, i] = maxval

        if test_cell is not None and \
                (test_cell == (j, i) or test_cell == (i, j)):

            M_local = matop.getSubMtx(M_rearranged, i, j)
            det_naive_local = np.linalg.det(M_local)


            if test_fill == 'min':
                M_rearranged[i, j] = M_rearranged[j, i] = interval[0]
            elif test_fill == 'max':
                M_rearranged[i, j] = M_rearranged[j, i] = interval[1]
            elif test_fill == 'rand':
                M_rearranged[i, j] = M_rearranged[j, i] = prng.uniform(*interval)
            if local_impact is not None:
                M_rand_local = matop.getSubMtx(M_rearranged, i, j)
                det_rand_local = np.linalg.det(M_rand_local)
                local_impact[i, j] += (det_naive_local - det_rand_local) /\
                    det_naive_local


        # print "inverse of matrix at iteration", (i, j)
        # print np.linalg.inv(matop.getSubMtx(M_rearranged, i, j))
        # oracle_prediction = oracle(M_original, inverting_order[i], inverting_order[j])
        #print '(%d, %d)  =  %.3g in range (%.3g, %.3g) >>>  DET = %.5g \t Oracle says %.3g' %\
            #(i, j, maxval, interval[0], interval[1],
             #np.linalg.det(matop.getSubMtx(M_rearranged, i, j)),
             #oracle_prediction)
        #assert np.linalg.det(matop.getSubMtx(M_rearranged, i, j)) >= 0
        #if oracle_prediction < interval[0] or oracle_prediction > interval[1]:
            #print 'TRUTH OUTSIDE INTERVAL! original position is', (inverting_order[i], inverting_order[j])

        #print M_rearranged
        #print M_rearranged[inverting_order, :][:, inverting_order]
        #print
        #print
        #print

        # set missing value to det maximizing value
    #print np.linalg.det(M_rearranged)
    if intervals is not None and local_impact is not None:
        return M_rearranged[inverting_order, :][:, inverting_order], intervals, local_impact
    else:
        return M_rearranged[inverting_order, :][:, inverting_order]
    return M_rearranged[inverting_order, :][:, inverting_order]


def single_run(n=10, num_diagonals=2, budget=0, strategy=None, seed=None):
    if seed is None:
        seed = np.random.randint(2000000000)
    print "PRNG seed:", seed
    prng = np.random.RandomState(seed)
    missing = -1

    M_original = generate_matrix(n, prng)
    print M_original
    print '============'
    M_partial, mask = partial_diagonal(M_original, num_diagonals)
    M_partial[mask > 0] = missing

    M_complete = complete(M_partial, M_original, mask, missing, oracle,
                                 budget, strategy)

    print
    print M_complete

    det_complete = np.linalg.det(M_complete)

    return det_complete


def multiple_runs(n=10, num_diagonals=2, max_budget=50, seed=None, repeat=10):
    if seed is None:
        seed = np.random.randint(2000000000)
    print "PRNG seed:", seed
    prng = np.random.RandomState(seed)
    missing = -1
    strategy_choice = [strategy_inner_order]
    results = []
    for _round in range(repeat):
        M_original = generate_matrix(n, prng)

        print M_original
        print '============'
        M_partial, mask = partial_diagonal(M_original, num_diagonals)
        M_partial[mask > 0] = missing

        #if _round == 0:
            #print mask
            #print

        M_naive = complete(M_partial.copy(), M_original,
                                  mask.copy(), missing, oracle,
                                  budget=0, strategy=None)
        det_naive = np.linalg.det(M_naive)
        print '============'
        print 'Naive', det_naive
        print '============'
        results = np.zeros((len(strategy_choice), min(mask.sum().sum(),
                                                      max_budget)))

        for budget in range(1, max_budget + 1):
            for strategy_index, strategy in enumerate(strategy_choice):
                M_complete = complete(M_partial.copy(), M_original,
                                      mask.copy(), missing, oracle,
                                      2, strategy)
                det_complete = np.linalg.det(M_complete)
                comparison = (det_naive - det_complete) / (1. * det_naive)
                results[strategy_index, budget - 1] = det_complete
                print '============'
    return results



def measure_cell_importance(n=10, num_diagonals=2, seed=None,
                            repeat=10, fill='rand', num_samples=1,
                            M_partial=None, M_original=None, mask=None):
    if seed is None:
        seed = np.random.randint(2000000000)
    print "PRNG seed:", seed
    prng = np.random.RandomState(seed)
    missing = -1
    for _round in range(repeat):
        if M_partial is None:
            M_original = generate_matrix(n, prng)

            #print M_original
            #print '============'
            M_partial, mask = partial_diagonal(M_original, num_diagonals)

        M_partial[mask > 0] = missing
        results = np.zeros(M_partial.shape)
        M_naive = complete(M_partial=M_partial.copy(), M_original=M_original,
                           mask=mask.copy(), missing=missing, oracle=oracle,
                           budget=0, strategy=None)
        print "Inverse of Naive completion"
        print np.linalg.inv(M_naive)
        det_naive = np.linalg.det(M_naive)
        print 'Naive determinant', det_naive
        print '============'
        intervals ={}
        local_impact = np.zeros(M_partial.shape)
        i, j = np.where(mask == 1)
        for missing_cell in zip(i, j):
            i, j = missing_cell
            if i > j:
                continue
            partial_res = 0
            for samples in range(num_samples):
                print "Working on rand cell", (i, j)
                M_test, intervals, local_impact = complete(M_partial=M_partial.copy(),
                                                           mask=mask.copy(),
                                                           missing=missing,
                                                           oracle=oracle,
                                                           M_original=M_original,
                                                           budget=0,
                                                           test_cell=missing_cell,
                                                           test_fill=fill,
                                                           prng=prng,
                                                           intervals=intervals,
                                                           local_impact=local_impact)
                R = matop.getSubMtx(M_test.copy(), i, j)
                R_inv = np.linalg.inv(R)
                X = matop.getSubMtx(M_naive.copy(), i, j)
                X_inv = np.linalg.inv(X)
                # print "X"
                # print X
                print
                print "Naive submatrix inverse"
                print X_inv
                print
                print "Naive inverse"
                print np.linalg.inv(M_naive)
                # print matop.getSubMtx(np.linalg.inv(M_naive), i, j)
                print
                print "Random submatrix inverse"
                print R_inv
                print
                print "Random inverse"
                print np.linalg.inv(M_test)
                # print matop.getSubMtx(np.linalg.inv(M_test), i, j)
                # print "Cofactor DetX", matop.cofactorDet(X, 0)
                # print "Cofactor DetR", matop.cofactorDet(R, 0)
                last_cof_X = ((-1.) ** (X.shape[1] - 1)) * X[0, -1] *\
                        matop.getMinor(X, 0, X.shape[1] - 1)
                last_cof_R = ((-1.) ** (R.shape[1] - 1)) * R[0, -1] *\
                        matop.getMinor(R, 0, R.shape[1] - 1)
                print "Cofactor last X", last_cof_X
                print "Cofactor last R", last_cof_R
                print "1-loc", np.linalg.det(R) / np.linalg.det(X)
                print  1 - ((last_cof_X - last_cof_R) / np.linalg.det(X))
                # print "the same?", matop.cofactorDet(R, 0)

                glob_impact = 1 - np.linalg.det(M_test) / np.linalg.det(M_naive)
                loc_impact = 1 - np.linalg.det(R) / np.linalg.det(X)
                # print matop.getMinor(R, 0, -1) / matop.getMinor(X, 0, -1)
                print "local impact:", loc_impact
                print "globl impact:", glob_impact
                r = np.atleast_2d(M_test[:, j]).T
                print r.shape
                lamd = np.linalg.inv(M_naive).T.dot(r)
                print lamd[j, 0]
                print
                print
                print X

                print R


                det_test = np.linalg.det(M_test)
                partial_res += det_naive - det_test
            local_impact[missing_cell] /= num_samples
            results[missing_cell] = partial_res / (det_naive * num_samples)
        print "Local impact of cells"
        print local_impact
        return results
