from __future__ import division
from numpy import log as ln, atleast_2d, logaddexp, all as npall, array
from numpy.linalg import slogdet, solve
from numpy.random import RandomState
from numpy import sqrt, exp, isnan, logical_and
from utils import get_largest_submatrix, log_subtract, count_nans
from graph_algorithms import (perfect_order, make_mask_graph,
                              min_deficiency_order)


def feasible_interval(M):
    """Given a matrix that is missing the top-right element, returns an
    completion interval that promises semi-positive definiteness and the
    maximum-determinant completion value.

    Returns
    =======
    maxdet_choice : float
        The choice for the missing element that results in a semi-positive
        definite matrix that has maximum determinant.
    interval : tuple
        The interval within which all choices result in a semi-positive
        definite competion.
    """
    n = M.shape[0]
    if n == 2:
        center = 0
        radius = sqrt(M[0, 0] * M[1, 1])
    else:
        B = M[0:n - 1, 0:n - 1]
        C = M[1:n, 1:n]
        A = M[1:n - 1, 1:n - 1]
        b = M[0, 1:n - 1]
        c = M[1:n - 1, n - 1]
        signA, logdetA = slogdet(A)
        signB, logdetB = slogdet(B)
        signC, logdetC = slogdet(C)
        if signA < 0 or signB < 0 or signC < 0:
            raise ValueError('A submatrix has negative determinant.')
        # radius = sqrt(detB * detC) / detA
        logradius = logdetB / 2
        logradius += logdetC / 2
        logradius -= logdetA
        radius = exp(logradius)
        A_inv_c = solve(A, c.T)
        center = b.dot(A_inv_c)
    interval = (center - radius, center + radius)
    return center, interval


def function_det(x, M):
    """Compute the determinant of M as a function of the element in the top
    right corner. Assumes the matrix is positive definite.
    """
    n = M.shape[0]
    if n > 2:
        B = M[0:n - 1, 0:n - 1]
        C = M[1:n, 1:n]
        A = M[1:n - 1, 1:n - 1]
        b = atleast_2d(M[0, 1:n - 1])
        c = atleast_2d(M[1:n - 1, n - 1])
        signA, logdetA = slogdet(A)
        signB, logdetB = slogdet(B)
        signC, logdetC = slogdet(C)
        logconstant = logdetB
        logconstant += logdetC
        logconstant -= 1 * logdetA
        constant = exp(logconstant)
        A_inv_c = solve(A, c.T)
        center = b.dot(A_inv_c)
        if npall(x != center):
            logdet = log_subtract(logconstant,
                                  logdetA + ln(abs(x - center)) * 2)
        else:
            logdets = []
            for x_i in x:
                if x != center:
                    logdets.append(log_subtract(logconstant,
                                                logdetA + ln(abs(x - center)) * 2))
                else:
                    logdets.append(logconstant)
            logdet = array(logdets)
        return exp(logdet)
        # return constant - exp(logdetA) * ((x - center) ** 2)
    else:
        center = 0
        constant = M[0, 0] * M[1, 1]
        return constant - (x - center) ** 2


def integral_function_det(x, y, M):
    """Computes the integral of `function_det` between `x` and `y`.
    """
    n = M.shape[0]
    if n > 2:
        B = M[0:n - 1, 0:n - 1]
        C = M[1:n, 1:n]
        A = M[1:n - 1, 1:n - 1]
        b = M[0, 1:n - 1]
        c = M[1:n - 1, n - 1]
        signA, logdetA = slogdet(A)
        signB, logdetB = slogdet(B)
        signC, logdetC = slogdet(C)
        logconstant = logdetB
        logconstant += logdetC
        logconstant -= 2 * logdetA  # later, we multiply with detA to drop the square
        constant = exp(logconstant)
        A_inv_c = solve(A, c.T)
        center = b.dot(A_inv_c)
    else:
        logdetA = 0
        center = 0
        constant = M[0, 0] * M[1, 1]
    integral = constant * (y - x)
    integral += (x - y) * (x ** 2 +
                           x * (y - 3 * center) +
                           y ** 2 -
                           3 * y * center +
                           3 * center ** 2) / 3
    integral *= exp(logdetA)
    return integral


def average_det(x, M):
    center, interval = feasible_interval(M)
    valid_x = x[logical_and(x >= interval[0], x <= interval[1])]
    if len(valid_x):
        return function_det(valid_x, M).mean()
    else:
        return integral_function_det(interval[0], interval[1], M) /\
            (interval[1] - interval[0])


def grone_order(partial, sample=False):
    order = min_deficiency_order(partial, sample=sample)
    G = make_mask_graph(partial)
    num_edges = len(G.edges())
    num_nodes = len(G)
    found_k = True
    add_edges = []
    # Compute the order in which we should complete the edges
    # as described in "Positive Definite Completions of Partial
    #  Hermitian Matrices" by Grone et al.
    while found_k and num_edges < num_nodes * (num_nodes - 1):
        found_k = False
        node_k_inv_index = len(order) - 1
        for node_k in order[::-1]:
            if len(G[node_k]) < num_nodes - 1:
                found_k = True
                break
            node_k_inv_index -= 1
        if found_k:
            for node_r in order[node_k_inv_index - 1::-1]:
                if node_k not in G[node_r]:
                    break
            if node_r >= node_k:
                node_r, node_k = node_k, node_r
            add_edges.append((node_r, node_k))
            G.add_edge(node_r, node_k)
            num_edges += 1
    # add_edges.reverse()
    return add_edges


def perfect_neighbors(partial):
    G = make_mask_graph(partial)
    order = []
    while G.number_of_edges() > 0:
        min_choice = (G.number_of_nodes(), None)
        for node in G:
            degree = len(G[node])
            if degree < min_choice[0]:
                min_choice = (degree, node)
        node = min_choice[1]
        degree = min_choice[0]
        neighbors = set(G[node]) - set([node])
        G.remove_node(node)
        for node_i in neighbors:
            for node_j in neighbors:
                if node_i != node_j:
                    G.add_edge(node_i, node_j)
        order.append(node)
    order.append(G.nodes()[0])
    G = make_mask_graph(partial)
    lex = {}
    for i, node in enumerate(order):
        lex[node] = i
    print lex
    add_edges = []
    num_edges = len(add_edges)
    while G.number_of_edges() < (len(G) ** 2 - len(G)) / 2:
        print 'edges', G.number_of_edges(), '/', (len(G) ** 2 - len(G)) / 2
        for i, node in enumerate(order):
            neighbors = set(G[node]) - set([node])
            for node_i in neighbors:
                if lex[node_i] < lex[node]:
                    continue
                for node_j in neighbors:
                    if lex[node_j] < lex[node]:
                        continue
                    if node_i != node_j and node_j not in G[node_i]:
                        if node_i > node_j:
                            node_i, node_j = node_j, node_i
                        G.add_edge(node_i, node_j)
                        add_edges.append((node_i, node_j))
        if len(add_edges) == num_edges:
            break
        num_edges = len(add_edges)
    return add_edges


def elimination_order(partial):
    G = make_mask_graph(partial)
    order = []
    while G.number_of_edges() > 0:
        min_choice = (G.number_of_nodes(), None)
        for node in G:
            degree = len(G[node])
            if degree < min_choice[0]:
                min_choice = (degree, node)
        node = min_choice[1]
        degree = min_choice[0]
        neighbors = set(G[node]) - set([node])
        G.remove_node(node)
        for node_i in neighbors:
            for node_j in neighbors:
                if node_i != node_j:
                    G.add_edge(node_i, node_j)
        order.append(node)

    order.reverse()
    add_edges = []
    for i, node_i in enumerate(order):
        for j in range(i):
            if i == j:
                continue
            node_j = order[j]
            if node_i > node_j:
                node_i, node_j = node_j, node_i
            add_edges.append((node_i, node_j))
    return add_edges


def maxdet_complete(partial):
    # starting_partial = partial.copy()
    add_edges = grone_order(partial)
    # G = make_mask_graph(partial)
    # clqtree = CliqueTree()
    # clqtree.from_graph(G)
    for edge in add_edges:
        i, j = edge
        to_complete = get_largest_submatrix(partial, i, j)
        if count_nans(to_complete) > 2:
            print 'nans maxdet', count_nans(to_complete)
        # if edge not in clqtree.insertable:
            # print 'Edge insertion makes graph non chordal:', (i, j)
            # starting_partial.dump('non_chordal.p')
            # raise TypeError
        guess, interval = feasible_interval(to_complete)
        to_complete[0, -1] = to_complete[-1, 0] = guess
        # print 'to_complete logdet', slogdet(to_complete)[1]
        if isnan(slogdet(to_complete)[1]):
            print 'nans', count_nans(to_complete)
        partial[i, j] = partial[j, i] = guess
        to_complete[0, -1] = to_complete[-1, 0] = guess
        # clqtree.add_edge(i, j)
    return partial


def rand_complete(partial, seed=None):
    res = _rand_complete(partial, seed)
    tried = 1
    while res is None and tried < 20:
        res = _rand_complete(partial, seed)
        tried += 1
    return res


def _rand_complete(partial, seed=None):
    prng = RandomState(seed)
    order = perfect_order(partial)
    G = make_mask_graph(partial)
    num_edges = len(G.edges())
    num_nodes = len(G)
    found_k = True
    add_edges = []
    # Compute the order in which we should complete the edges
    # as described in "Positive Definite Completions of Partial
    #  Hermitian Matrices" by Grone et al.
    while found_k and num_edges < num_nodes * (num_nodes - 1):
        found_k = False
        node_k_inv_index = len(order) - 1
        for node_k in order[::-1]:
            if len(G[node_k]) < num_nodes - 1:
                found_k = True
                break
            node_k_inv_index -= 1
        if found_k:
            for node_r in order[node_k_inv_index - 1::-1]:
                if node_k not in G[node_r]:
                    break
            add_edges.append((node_r, node_k))
            G.add_edge(node_r, node_k)
            num_edges += 1
    add_edges.reverse()

    while add_edges:
        i, j = add_edges.pop()
        to_complete = get_largest_submatrix(partial, i, j)
        fill_info = feasible_interval(to_complete)
        if fill_info is None:
            print 'stopping'
            return None
        guess, interval = fill_info
        # picked = prng.uniform(interval[0], interval[1])
        incr = (interval[1] - interval[0]) * 0.25
        picked = prng.uniform(interval[0] + incr, interval[1] - incr)
        # picked = prng.normal(guess, (interval[1] - interval[0]) / 10)

        partial[i, j] = partial[j, i] = picked
    return partial
