import numpy as np
from numpy import log as ln, exp
from sklearn.cluster import KMeans
import matplotlib.pyplot as plt
from graph_algorithms import make_mask_graph

MAXINT = 1e8


def is_symmetric(M):
    return (M.T - M < 1e-10).all()


def is_pd(M):
    return np.all(np.linalg.eigvals(M) > -1e-10)


def plot_boolean_matrix(mtx, mask=True, figsize=(8, 8)):
    plt.figure(figsize=figsize)
    cmin = np.amin(mtx[~np.isnan(mtx)])
    cmax = np.amax(mtx[~np.isnan(mtx)])
    data = (mtx * 1.0 - cmin) / (cmax - cmin)
    plt.imshow(data, cmap=plt.cm.gray, interpolation='nearest')
    if mask:
        mask_mtx = np.zeros(mtx.shape)
        mask_mtx[np.isnan(mtx)] = 1
        plt.imshow(mask_mtx, cmap=plt.cm.YlOrBr, alpha=0.25,
                   interpolation='nearest')


def rearrange(mtx, new_indices, return_inverse=False):
    inverse = [new_indices.index(i) for i in range(mtx.shape[0])]
    if return_inverse:
        return mtx[new_indices].T[new_indices], inverse
    return mtx[new_indices].T[new_indices]


def disagreement_distance(left, right):
    if len(left) != len(right):
        raise ValueError('The two lists should have the same left')
    disagreements = 0
    for i in range(len(left)):
        for j in range(i + 1, len(right)):
            if (left[i] == left[j] and right[i] != right[j]) or\
                    (left[i] != left[j] and right[i] == right[j]):
                disagreements += 1
    return disagreements


def compute_kmeans(mtx, n_clusters):
    estimator = KMeans(init='k-means++', n_clusters=n_clusters,
                       n_init=200)
    estimator.fit(mtx)
    return estimator.labels_


def evaluate_kmeans(candidate, mtx, n_clusters, repeats=20):
    candidate_kmeans = []
    ground_labels = compute_kmeans(mtx, n_clusters)
    for tries in range(repeats):
        candidate_labels = compute_kmeans(candidate, n_clusters)
        candidate_kmeans.append(disagreement_distance(candidate_labels,
                                                      ground_labels))

    n = mtx.shape[0]
    score = np.mean(candidate_kmeans) / (n * (n - 1) / 2) * 100
    return score


def get_largest_submatrix(partial, i, j, return_indices=False):
    if i > j:
        i, j = j, i
    cols = np.where(~np.isnan(partial[i, :]))[0]
    rows = np.where(~np.isnan(partial[j, :]))[0]
    keep = [i]
    for x in sorted(set(cols).intersection(set(rows))):
        keep.append(x)

    keep.append(j)
    nans = set()
    temp = partial[keep, :][:, keep]
    w = np.where(np.isnan(temp))
    for u, v in zip(w[0], w[1]):
        if u > 0 and u < len(keep) - 1:
            nans.add(keep[u])
    keep = [x for x in keep if x not in nans]
    to_return = partial[keep, :][:, keep]
    if count_nans(to_return) > 2:
        raise ValueError('wrong submatrix')
    if return_indices:
        return partial[keep, :][:, keep], keep
    else:
        return partial[keep, :][:, keep]


def log_subtract(x, y):
    if np.any(x < y):
        raise ValueError('Cannot compute the log of non-positive number')
    res_old = x + np.log1p(-1 * np.exp(y - x))
    res = x + log1mexp(y - x)
    assert np.allclose(res_old, res)
    # res = np.log(np.exp(x - y) + 1) + y
    return res


def log1mexp(a):
    """Compute the value of log(1 - exp(a)) accurately.

        We refer to "Accurately computing log(1-exp(-|a|)) assessed by th Rmpfr
        package" by Martin Maechler.
    """
    if np.any(a > 0):
        raise ValueError('Cannot compute the log(1-exp(a)) when a > 0.')
    if not isinstance(a, np.ndarray) and not isinstance(a, list):
        if a < - ln(2):
            return np.log1p(-exp(a))
        else:
            return ln(-np.expm1(a))
    else:
        res = []
        for ai in a.reshape(-1):
            if ai < - ln(2):
                res.append(np.log1p(-exp(ai)))
            else:
                res.append(ln(-np.expm1(ai)))
        return np.array(res)


def count_nans(M):
    return np.isnan(M).sum()


def is_perfect_elimination_order(order, partial):
    G = make_mask_graph(partial)
    lex = {}
    for i, node in enumerate(order):
        lex[node] = i
    for node in order:
        post_neighbors = [v for v in G[node] if lex[v] > lex[node]]
        if len(post_neighbors) > 1:
            # print node, '->', post_neighbors
            for u in post_neighbors:
                for v in post_neighbors:
                    if u > v:
                        if not (u in G[v] or v in G[u]):
                            # print 'oops', u, ':', G[u].keys(), v, ':', G[v].keys()
                            return False
    return True


def is_perfect_elimination_order_fast(order, partial=None, G=None):
    """Checks if the order is a perfect elimination order for the mask graph.

    Uses the algorithm described in [1]

    References
    ----------
    [1] M. C. Golumbic, Algorithmic Graph Theory and
        Perfect Graphs. Elsevier, 2014.
    """
    if G is None:
        if partial is None:
            raise ValueError('Either the mask graph or the partial matrix are needed as input.')
        G = make_mask_graph(partial)
    order_position = {v: i for i, v in enumerate(order)}
    inv_order_position = {i: v for i, v in enumerate(order)}
    downstream = {v: {u for u in G[v] if order_position[v] < order_position[u]}
                  for v in G.nodes()}
    required_adjacent = {v: set() for v in G.nodes()}

    for i, v in enumerate(order[:-1]):
        downstream_v = downstream[v]
        if downstream_v:
            min_downstream_position = min([order_position[x]
                                           for x in downstream_v])
            u = inv_order_position[min_downstream_position]
            required_adjacent[u] |= downstream_v - {u}
        if len(required_adjacent[v] - downstream_v) > 0:
            return False
    return True
